 
Inductive type : Set :=
  tV: nat ->  type
 | arrow: type -> type ->  type.
 
Inductive cl : Set :=
  cK: cl
 | cS: cl
 | V: nat -> type ->  cl
 | App: cl -> cl ->  cl.
 
Inductive wt : cl -> type ->  Prop :=
  type_S:
    forall (a b c : type),
     wt cS (arrow (arrow a (arrow b c)) (arrow (arrow a b) (arrow a c)))
 | type_K: forall (a b : type),  wt cK (arrow a (arrow b a))
 | type_v: forall (n : nat) (t : type),  wt (V n t) t
 | type_app:
     forall (a b : type) (e1 e2 : cl),
     wt e1 (arrow a b) -> wt e2 a ->  wt (App e1 e2) b.
Hint Resolve type_S type_K type_v type_app.
 
Inductive red : cl -> cl ->  Prop :=
  red_K: forall (e1 e2 : cl),  red (App (App cK e1) e2) e1
 | red_S: forall (e1 e2 e3 : cl),
      red (App (App (App cS e1) e2) e3) (App (App e1 e3) (App e2 e3))
 | red_1: forall (e1 e2 e3 : cl), red e1 e3 ->  red (App e1 e2) (App e3 e2)
 | red_2: forall (e1 e2 e3 : cl), red e2 e3 ->  red (App e1 e2) (App e1 e3).
Hint Resolve red_K red_S red_1 red_2.
 
Ltac someInversion :=
  intros; match goal with | id:_ |- _ => inversion id; auto; fail end.
 
Definition red' (x y : cl) := red y x.
Hint Unfold red'.
 
Fixpoint reducible (t : type) (e: cl) {struct t} : Prop :=
 match t with
   tV n =>  wt e (tV n) /\ Acc red' e
  | arrow t1 t2 =>
      wt e (arrow t1 t2) /\
      (forall (e1 : cl), reducible t1 e1 ->  reducible t2 (App e e1))
 end.
 
Theorem reducible_typed: forall (e : cl) (t : type), reducible t e ->  wt e t.
intros e t; case t; simpl; intuition.
Qed.
Hint Resolve reducible_typed.
 
Ltac type_tac :=
match goal with
| id:wt ?e1 (arrow ?t' ?t) |- wt (App ?e1 ?e2) ?t =>
apply type_app with t'; type_tac
| id:wt ?e (arrow ?t' ?t) |- wt (App _ (App ?e _)) _ =>
apply type_app with t; type_tac
| id:reducible ?t ?e |- wt (App ?e1 ?e) _ => apply type_app with t; type_tac
| id:wt ?e ?t'' |- wt (App ?e1 ?e) _ => apply type_app with t''; type_tac
| |- _ => auto end.
 
Theorem srt: forall (e e' : cl) (t : type), wt e t -> red e e' ->  wt e' t.
intros e e' t H; generalize e'; elim H; clear H e e' t; (try someInversion).
intros a b e1 e2 H H0 H1 H2 e' H3; inversion H3.
subst; inversion H.
subst; match goal with | id:wt cK _ |- _ => inversion id end.
subst; auto.
subst; inversion H.
subst; match goal with | id:wt (App cS _) _ |- _ => inversion id end.
subst; match goal with
       | id:wt cS _ |- _ => inversion id end.
subst; type_tac.
subst; type_tac.
subst; type_tac.
Qed.
Hint Resolve srt.
 
Inductive neutral : cl ->  Prop :=
  neutral_4: forall (e1 e2 e3 e4 : cl),  neutral (App (App (App e1 e2) e3) e4)
 | neutral_K: forall (e1 e2 : cl),  neutral (App (App cK e1) e2)
 | neutral_3:
     forall (n : nat) (t : type) (e1 e2 : cl),
      neutral (App (App (V n t) e1) e2)
 | neutral_2: forall (n : nat) (t : type) (e1 : cl),  neutral (App (V n t) e1)
 | neutral_1: forall (n : nat) (t : type),  neutral (V n t).
Hint Resolve neutral_4 neutral_K neutral_3 neutral_2 neutral_1.
 
Section multiple_acc.
Variables (A : Set) (R : A -> A ->  Prop).
 
Theorem accElim2:
 forall (P : A -> A ->  Prop),
 (forall (x y : A),
  (forall (t : A), R t x ->  P t y) ->
  (forall (t : A), R t y ->  P x t) ->  P x y) ->
 forall (x y : A), Acc R x -> Acc R y ->  P x y.
intros P H x y H0; generalize y; clear y; elim H0.
intros x0 H1 H2 y H3; elim H3.
intros x1 H4 H5.
apply H; auto.
intros t H6; apply H2; auto.
apply Acc_intro; auto.
Qed.
 
Theorem accElim3:
 forall (P : A -> A -> A ->  Prop),
 (forall (x y z : A),
  (forall (t : A), R t x ->  P t y z) ->
  (forall (t : A), R t y ->  P x t z) ->
  (forall (t : A), R t z ->  P x y t) ->  P x y z) ->
 forall (x y z : A), Acc R x -> Acc R y -> Acc R z ->  P x y z.
intros P H x y z H0; generalize y z; clear y z; elim H0.
intros x0 H1 H2 y z H3; generalize z; clear z; elim H3.
intros x1 H4 H5 z H6; elim H6.
intros x2 H7 H8; apply H; auto.
intros t H9; apply H2; auto.
apply Acc_intro; auto.
apply Acc_intro; auto.
intros t H9; apply H5; auto.
apply Acc_intro; auto.
Qed.
 
End multiple_acc.
 
Theorem l1_aux:
 forall (x : cl),
 Acc red' x -> forall (e1 e2 : cl), x = App e1 e2 ->  Acc red' e1.
intros x H; elim H.
intros x0 H0 H1 e1 e2 H2.
apply Acc_intro.
intros y H3.
apply H1 with (App y e2) e2; auto.
rewrite H2; auto.
Qed.
 
Theorem l1: forall (e1 e2 : cl), Acc red' (App e1 e2) ->  Acc red' e1.
intros; eapply l1_aux; eauto.
Qed.
 
Theorem l1'_aux:
 forall (x : cl),
 Acc red' x -> forall (e1 e2 : cl), x = App e1 e2 ->  Acc red' e2.
intros x H; elim H.
intros x0 H0 H1 e1 e2 H2.
apply Acc_intro.
intros y H3.
apply H1 with (App e1 y) e1; auto.
rewrite H2; auto.
Qed.
 
Theorem l1': forall (e1 e2 : cl), Acc red' (App e1 e2) ->  Acc red' e2.
intros; eapply l1'_aux; eauto.
Qed.
 
Theorem reducible_base_acc:
 forall (n : nat) (e : cl), reducible (tV n) e ->  Acc red' e.
intros n e [H H']; auto.
Qed.
 
Theorem reducible_base_red_reducible:
 forall (n : nat) (e : cl),
 reducible (tV n) e -> forall (e' : cl), red e e' ->  reducible (tV n) e'.
intros n e [H H'] e' Hr; split; eauto.
eapply Acc_inv; eauto.
Qed.
 
Theorem reducible_base_all_red:
 forall (n : nat) (e : cl), wt e (tV n) ->
 (forall (e' : cl), red e e' ->  reducible (tV n) e') ->  reducible (tV n) e.
intros n e H H1; split; auto.
apply Acc_intro.
intros y H2; elim (H1 y); auto.
Qed.
Hint Resolve
 reducible_base_acc reducible_base_red_reducible reducible_base_all_red.
 
Theorem reducible_arr_acc:
 forall (t1 t2 : type),
 (forall (e : cl), reducible t2 e ->  Acc red' e) ->
 (forall (e : cl),  wt e t1 -> neutral e ->
  (forall (e' : cl), red e e' ->  reducible t1 e') ->  reducible t1 e) ->
 forall (e : cl), reducible (arrow t1 t2) e ->  Acc red' e.
simpl; intros t1 t2 H H1 e [H0 H'].
apply l1 with (V 0 t1).
apply H; apply H'; apply H1; auto.
intros; someInversion.
Qed.
 
Theorem reducible_arr_red_reducible:
 forall (t1 t2 : type),
 (forall (e : cl),
  reducible t2 e -> forall (e' : cl), red e e' ->  reducible t2 e') ->
 forall (e : cl),
 reducible (arrow t1 t2) e ->
 forall (e' : cl), red e e' ->  reducible (arrow t1 t2) e'.
simpl; intros t1 t2 H e [H0 H0'] e' H1; split; eauto.
Qed.
Hint Resolve reducible_arr_acc reducible_arr_red_reducible.
 
Theorem neutral_app: forall (e1 e2 : cl), neutral e1 ->  neutral (App e1 e2).
intros e1 e2 H; inversion H; auto.
Qed.
Hint Resolve neutral_app.
 
Theorem reducible_arr_all_red:
 forall (t1 t2 : type),
 (forall (e : cl),
  reducible t1 e -> forall (e' : cl), red e e' ->  reducible t1 e') ->
 (forall (e : cl), reducible t1 e ->  Acc red' e) ->
 (forall (e : cl),
  wt e t2 -> neutral e ->
  (forall (e' : cl), red e e' ->  reducible t2 e') ->  reducible t2 e) ->
 forall (e : cl),
 wt e (arrow t1 t2) -> neutral e ->
 (forall (e' : cl), red e e' ->  reducible (arrow t1 t2) e') ->
  reducible (arrow t1 t2) e.
intros t1 t2 H H0 H'2 e H1 H2 H3; (try assumption).
split; [auto | intros e0 H4].
generalize (H0 e0 H4); intros H5; generalize H4; clear H4.
elim H5; clear H5 e0.
intros e0 H4 H5 H6; (try assumption).
apply H'2; [idtac | auto | idtac].
type_tac.
intros e' H7.
generalize H2; clear H2; inversion H7; intros;
 (try match goal with
      | id:neutral _ |- _ => inversion id; fail end).
elim (H3 e3); eauto.
apply H5; [exact H10 | apply H with e0; auto].
Qed.
 
Theorem main_prop:
 forall (t : type) (e : cl),
  (reducible t e ->  Acc red' e) /\
  ((reducible t e -> forall (e' : cl), red e e' ->  reducible t e') /\
   (wt e t -> neutral e ->
    (forall (e' : cl), red e e' ->  reducible t e') ->  reducible t e)).
intros t; elim t.
split; [idtac | split]; eauto.
intros t0 H t1 H0 e;
 (split;
   [intros H1 |
    split; [intros H1; intros e'; intros H2; (try assumption) | idtac]]).
apply reducible_arr_acc with t0 t1; auto.
intros e0; elim (H0 e0); auto.
intros e0; elim (H e0); intros H' [H'1 H'2]; auto.
apply reducible_arr_red_reducible with e; auto.
intros e0; elim (H0 e0); intros H' [H'1 H'2]; auto.
apply reducible_arr_all_red.
intros e0; elim (H e0); intros H' [H'1 H'2]; auto.
intros e0; elim (H e0); intros H' [H'1 H'2]; auto.
intros e0; elim (H0 e0); intros H' [H'1 H'2]; auto.
Qed.
 
Theorem main_prop1: forall (t : type) (e : cl), reducible t e ->  Acc red' e.
intros t e; elim (main_prop t e); intros H1 [H2 H3]; auto.
Qed.
Hint Resolve main_prop1.
 
Theorem main_prop2:
 forall (t : type) (e : cl),
 reducible t e -> forall (e' : cl), red e e' ->  reducible t e'.
intros t e; elim (main_prop t e); intros H1 [H2 H3]; auto.
Qed.
 
Theorem main_prop3:
 forall (t : type) (e : cl),
 wt e t -> neutral e ->
 (forall (e' : cl), red e e' ->  reducible t e') ->  reducible t e.
intros t e; elim (main_prop t e); intros H1 [H2 H3]; auto.
Qed.
 
Theorem V_reducible: forall (t : type) (n : nat),  reducible t (V n t).
intros t n.
apply main_prop3; auto.
intros e' H'; inversion H'.
Qed.
Hint Resolve V_reducible.
 
Theorem decompose_K_redex:
 forall (e1 e2 : cl) (P : cl ->  Prop),
 P e1 ->
 (forall (e : cl), red e1 e ->  P (App (App cK e) e2)) ->
 (forall (e : cl), red e2 e ->  P (App (App cK e1) e)) ->
 forall (e : cl), red (App (App cK e1) e2) e ->  P e.
intros e1 e2 P H H0 H1 e H2; inversion H2; auto.
rewrite <- H3; auto.
inversion H6; auto.
inversion H10.
Qed.
 
Theorem decompose_S_redex:
 forall (e1 e2 e3 : cl) (P : cl ->  Prop),
 P (App (App e1 e3) (App e2 e3)) ->
 (forall (e : cl), red e1 e ->  P (App (App (App cS e) e2) e3)) ->
 (forall (e : cl), red e2 e ->  P (App (App (App cS e1) e) e3)) ->
 (forall (e : cl), red e3 e ->  P (App (App (App cS e1) e2) e)) ->
 forall (e : cl), red (App (App (App cS e1) e2) e3) e ->  P e.
intros e1 e2 e3 P H H0 H1 H2 e H3; inversion H3; auto.
generalize H0 H1; clear H0 H1.
inversion H7; auto.
inversion H9; auto.
inversion H13.
Qed.
 
Theorem l2:
 forall (e1 e2 : cl) (t t' t'' : type),
 wt e1 (arrow t'' (arrow t' t)) -> wt e2 (arrow t'' t') ->
 (forall (e3 : cl),
  reducible t'' e3 ->  reducible t (App (App e1 e3) (App e2 e3))) ->
  reducible (arrow t'' t) (App (App cS e1) e2).
simpl.
intros e1 e2 t t' t'' H H0 H1; split.
type_tac.
intros e H4.
cut (Acc red' e1).
intros H2; cut (Acc red' e2).
intros H3; generalize (main_prop1 _ _ H4).
intros H5; generalize H4 H1 H0 H; clear H H0 H1 H4.
apply accElim3 with ( 2 := H2 ) ( 3 := H3 ) ( 4 := H5 ); clear H5 H3 H2 e e1 e2.
intros e1 e2 e Hrec1 Hrec2 Hrec3 H4 H1 H0 H.
apply main_prop3; auto.
type_tac.
intros e' H2.
elim H2  using decompose_S_redex; auto.
intros e0 H3.
apply Hrec1; eauto.
intros e3 H5; eapply main_prop2 with (App (App e1 e3) (App e2 e3)); eauto.
intros e0 H3.
apply Hrec2; eauto.
intros e3 H5; eapply main_prop2 with (App (App e1 e3) (App e2 e3)); eauto.
intros e0 H3.
apply Hrec3; eauto.
apply main_prop2 with e; auto.
eapply l1; eapply l1'; eauto.
eapply l1; eapply l1; eauto.
Qed.
 
Theorem l3:
 forall (e1 : cl) (t : type),
 wt e1 t -> reducible t e1 ->
 forall (t' : type),  reducible (arrow t' t) (App cK e1).
simpl.
intros e1 t H H0 t'; split.
type_tac.
intros e H1; cut (Acc red' e1).
intros H2; cut (Acc red' e).
intros H3; generalize H H0 H1; apply accElim2 with ( 3 := H2 ) ( 2 := H3 ).
clear H3 H2 H1 H0 H e1 e.
intros e e1 Hrec1 Hrec2 H1 H2 H3.
apply main_prop3; auto.
type_tac.
intros e' H; elim H  using decompose_K_redex; auto.
intros e0 H0; apply Hrec2; eauto.
eapply main_prop2; eauto.
intros e0 H0; apply Hrec1; eauto.
apply main_prop2 with e; auto.
apply main_prop1 with t'; auto.
apply main_prop1 with t; auto.
Qed.
Hint Resolve l3.
 
Theorem l4: forall (t t' : type),  reducible (arrow t (arrow t' t)) cK.
intros; split; auto.
Qed.
Hint Resolve l4.
 
Theorem reducible_app:
 forall (t1 t2 : type) (e1 e2 : cl),
 reducible (arrow t1 t2) e1 -> reducible t1 e2 ->  reducible t2 (App e1 e2).
simpl; intros t1 t2 e1 e2 [H H']; auto.
Qed.
Hint Resolve reducible_app.
 
Theorem l7:
 forall (t t' t'' : type),
  reducible
   (arrow (arrow t (arrow t' t'')) (arrow (arrow t t') (arrow t t''))) cS.
intros t t' t''.
split; auto.
intros e1 H; split; auto.
type_tac.
intros e2 H0; apply l2 with t'.
apply reducible_typed; auto.
apply reducible_typed; auto.
eauto.
Qed.
Hint Unfold reducible.
Hint Resolve l7.
 
Theorem l6': forall (e : cl) (t : type), wt e t ->  reducible t e.
intros e t H; elim H; auto.
intros t1 t2 e1 e2 wte1 IHe1 wte2 IHe2; elim IHe1; auto.
Qed.
Hint Resolve l6'.
 
Theorem strong_normalization:
  forall (e : cl) (t : type), wt e t ->  Acc red' e.
intros e t H; apply main_prop1 with t; auto.
Qed.
